# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import traceback
import sys


class BaseCommand(object):
    def _ssh_data(self, sshc, command):
        res = ''
        try:
            stdin, stdout, stderr = sshc.exec_command(command)
            res = str(stdout.read())
        except Exception:
            print(traceback.format_exc(), file = sys.stderr)

        return res

    def _ssh_data_with_header(self, sshc, command):
        data = self._ssh_data(sshc, command)
        res = []

        try:
            if ' 0 ' in data:
                res = data.partition(' 0 ')[2].split('\\r\\n\\r\\n')[:-1]
                res = list(map(lambda y: self._parse_data(y), res))
        except Exception:
            print(traceback.format_exc(), file = sys.stderr)

        return res

    def _parse_data(self, data):
        split_data = data.replace(' \\r\\n ', '').replace('\'', '').split('=')
        return dict(zip(list(map(lambda x: x.rpartition(' ')[-1].strip().replace('\"', ''), split_data[:-1])), \
                        list(map(lambda x: x.rpartition(' ')[0].strip().replace('\"', ''), \
                                 split_data[1:-1])) + [split_data[-1].strip().replace('\"', '')]))

    def run_ssh(self, data):
        raise NotImplementedError



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import traceback
import re
import os
import sys

from commands.basecommand import BaseCommand
from nvd import CVEValidator

CVES_PATH = './assets/mikrotik_cpe_match.json'


class Version(BaseCommand):
    def __init__(self):
        self.__name__ = 'Version'

    def run_ssh(self, sshc):
        version = ''
        data = self._ssh_data(sshc, ':put [/system resource get version]')

        try:
            version_reg = re.search(r'([\d\.]+)', data)

            if version_reg:
                version = version_reg.group(1)

        except Exception:
            print(traceback.format_exc())

        sus, recommendation = self.check_results_ssh(version)

        return {'raw_data': version,
                'suspicious': sus,
                'recommendation': recommendation}

    def check_results_ssh(self, version):
        sus_version = []
        recommendation = []

        try:
            if version and os.path.isfile(CVES_PATH):
                cve = CVEValidator(CVES_PATH)
                ver_cves = cve.check_version(version)
                if ver_cves:
                    sus_version = ver_cves
                    recommendation.append(f'RouterOS version: {version} is vulnerable to CVE(s). Upgrade to the latest version. (The CVEs list is from NVD)')
        except Exception:
            print(traceback.format_exc(), file = sys.stderr)

        return sus_version, recommendation






# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import traceback
import re
import sys

from commands.basecommand import BaseCommand


class Scheduler(BaseCommand):
    def __init__(self):
        self.__name__ = 'Scheduler'

    def run_ssh(self, sshc):
        res = self._ssh_data_with_header(sshc, '/system scheduler print detail')
        sus_dns, recommendation = self.check_results_ssh(res)

        return {'raw_data': res,
                'suspicious': sus_dns,
                'recommendation': recommendation}

    def check_results_ssh(self, res):
        sus_tasks = []
        recommendation = []

        try:
            for item in res:
                if (re.match(r'u\d+$', item['name'].lower())) or (('/tool fetch' in item['on-event']) or
                                                                  ('url' in item['on-event']) or ('http' in item['on-event'])):
                    sus_tasks.append(f'Task name: {item["name"]}, executes: {item["on-event"]} - severity: high')
        except Exception:
            print(traceback.format_exc(), file = sys.stderr)

        return sus_tasks, recommendation












# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import argparse
import json
import traceback
import paramiko
import os
import datetime
import sys
import re

from commands.fwnat import FWNat
from commands.dns import DNS
from commands.files import Files
from commands.fwrules import FW
from commands.ports import Ports
from commands.proxy import Proxy
from commands.scheduler import Scheduler
from commands.socks import Socks
from commands.users import Users
from commands.version import Version
import query_nvd

CVES_PATH = './assets/mikrotik_cpe_match.json'


def main(args):
    all_data = {}
    commands = [Version(), Scheduler(), Files(), FWNat(), Proxy(), Socks(), DNS(), Users(), Ports(), FW()]

    if args.update or is_cves_file_updated():
        update_cves()

    print(f'** Mikrotik ip address: {args.ip}\n')

    with paramiko.SSHClient() as ssh_client:
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh_client.connect(hostname=args.ip, port=args.port, username=args.userName, password=args.password,
        look_for_keys=False, allow_agent=False)

        for command in commands:
            res = command.run_ssh(ssh_client)
            all_data[command.__name__] = res
    
        if args.J:
            print(json.dumps(all_data, indent=4))
        else:
            print_txt_results(all_data, args.concise)

def print_txt_results(res, concise):
    for command in res:
        if (not concise and res[command]["raw_data"]) or res[command]["recommendation"] or res[command]["suspicious"]:
            print(f'{command}:')
            for item in res[command]:
                if concise and item != "recommendation" and item != "suspicious":
                        continue
                if res[command][item]:
                    print(f'\t{item}:')
                    if type(res[command][item]) == list:
                        data = '\n\t\t'.join(json.dumps(i) for i in res[command][item])
                    else:
                        data = res[command][item]
                    print(f'\t\t{data}')

def is_cves_file_updated():
    if os.path.isfile(CVES_PATH):
        month_ago = datetime.datetime.now() - datetime.timedelta(30)
        modification_date = datetime.datetime.fromtimestamp(os.path.getmtime(CVES_PATH))
        need_update = month_ago > modification_date
        if need_update:
            print("** The CVE data file hasn't been update in the last month", file = sys.stderr)
    else:
        print("** There is no CVE data file", file = sys.stderr)
        need_update = True

    return need_update

def update_cves():
    print(f'** Trying to update the CVE json file', file = sys.stderr)
    try:
        ci = query_nvd.CVEsInterface()

        ver_cves = ci.nist_api('mikrotik', 'routeros')
        strip_no_routeros_version(ver_cves)

        if ver_cves:
            with open('./assets/mikrotik_cpe_match.json', 'w') as fjson:
                fjson.write(json.dumps(ver_cves, indent=4))
            print("** The CVE update process succeeded", file=sys.stderr)
        else:
            print(
                "** The CVE update process failed, if an older version of the json file is available, the program will use it, else the program won't return CVE data.",
                file=sys.stderr)
    except Exception:
        print(f"** The CVE update process failed, if an older version of the json file is available, the program will use it, else the program won't return CVE data.\n"
              f"   The exception message:\n {traceback.format_exc()}", file = sys.stderr)

def strip_no_routeros_version(cves):
    for cve in cves:
        for versions in cves[cve]:
            for ver in list(versions):
                if not is_routeros_version(versions[ver]):
                    versions.pop(ver)

def is_routeros_version(ver):
    return re.match(r'\d{1,2}\.\d{1,2}(\.\d{1,2})?', ver)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--ip', help='The tested Mikrotik IP address', required=True)
    parser.add_argument('-p', '--port', help='The tested Mikrotik SSH port', default='22')
    parser.add_argument('-u', '--userName', help='User name with admin Permissions', required=True)
    parser.add_argument('-ps', '--password', help='The password of the given user name', default='')
    parser.add_argument('-J', help='Print the results as json format', action='store_true')
    parser.add_argument('-concise', help='Print out only suspicious items and recommendations', action='store_true')
    parser.add_argument('-update', help='Update the CVE Json file', action='store_true')
    args = parser.parse_args()

    main(args)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import json

class Comparison(object):
    SMALLER = -1
    SAME = 0
    BIGGER = 1
    

class CVEValidator(object):
    def __init__(self, jsonfname):
        with open(jsonfname, 'r') as fjson:
            self._all_cpe_match_data = json.loads(fjson.read())
    
    def _compare_3_section_version(self, version, version_to_compare_to):
        va_splitted = version.split('.')
        vb_splitted = version_to_compare_to.split('.')
        
        comparison = Comparison.SAME
        
        for index in range(3):
            a = 0
            
            if len(va_splitted) > index:
                a = int(va_splitted[index])
                
            b = 0
            
            if len(vb_splitted) > index:
                b = int(vb_splitted[index])
            
            if a == b:
                continue
            elif a < b:
                comparison = Comparison.SMALLER
                break
            else:
                comparison = Comparison.BIGGER
                break
                
        return comparison
    
    def check_version(self, version):
        res = []
        
        for cve in self._all_cpe_match_data:
            for match_ranges in self._all_cpe_match_data[cve]:
                if 'start_including' in match_ranges:
                    if self._compare_3_section_version(version, \
                                                       match_ranges['start_including']) >= Comparison.SAME:
                        if 'end_including' in match_ranges:
                            if self._compare_3_section_version(version, \
                                                       match_ranges['end_including']) <= Comparison.SAME:
                                res.append(cve)
                        elif 'end_excluding' in match_ranges:
                            if self._compare_3_section_version(version, \
                                                       match_ranges['end_excluding']) < Comparison.SAME:
                                res.append(cve)
                        else:
                            res.append(cve)
                elif 'end_including' in match_ranges:
                    if self._compare_3_section_version(version, \
                                               match_ranges['end_including']) <= Comparison.SAME:
                        res.append(cve)
                elif 'start_excluding' in match_ranges:
                    if self._compare_3_section_version(version, \
                                                       match_ranges['start_excluding']) > Comparison.SAME:
                        if 'end_including' in match_ranges:
                            if self._compare_3_section_version(version, \
                                                       match_ranges['end_including']) <= Comparison.SAME:
                                res.append(cve)
                        elif 'end_excluding' in match_ranges:
                            if self._compare_3_section_version(version, \
                                                       match_ranges['end_excluding']) < Comparison.SAME:
                                res.append(cve)
                        else:
                            res.append(cve)
                elif 'end_excluding' in match_ranges:
                    if self._compare_3_section_version(version, \
                                               match_ranges['end_excluding']) < Comparison.SAME:
                        res.append(cve)
                elif 'exact' in match_ranges:
                    if self._compare_3_section_version(version, \
                                               match_ranges['exact']) == Comparison.SAME:
                        res.append(cve)
                        
        return list(set(res))


# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import traceback
import sys

from commands.basecommand import BaseCommand


class Users(BaseCommand):
    def __init__(self):
        self.__name__ = 'Users'

    def run_ssh(self, sshc):
        res = self._ssh_data_with_header(sshc, '/user print detail')
        sus_dns, recommendation = self.check_results_ssh(res)

        return {'raw_data': res,
                'suspicious': sus_dns,
                'recommendation': recommendation}

    def check_results_ssh(self, res):
        sus_users = []
        recommendation = []

        try:
            for item in res:
                if (item['name'] == 'admin') and (item['group'] == 'full'):
                    recommendation.append(
                        'You are using the default "admin" user name- create new user in "full" group with a unique name, '
                        'and delete the admin user')
                if item['address'] == '':
                    recommendation.append(f'Add allowed ip address to user: {item["name"]}, '
                                          f'to be the only address it can login from')
        except Exception:
            print(traceback.format_exc(), file = sys.stderr)

        return sus_users, recommendation










# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

import requests
import collections
import sys
from retry import retry


class NvdApiError(Exception):
    """Exception raised for errors in the API request """
    def __init__(self, params,  msg):
        self.message = f"Received an error during the API process with the following params: {params}. The API message: {msg}"
        super().__init__(self.message)

class hashabledict(dict):
    def __key(self):
        summ = []
        for k in sorted(self):
            v = self[k]
            if type(v) == dict:
                summ.append((k, hashabledict(v)))
            elif type(v) == list:
                summ.append((k, tuple(sorted(str(v)))))
            else:
                summ.append((k, v))
        return tuple(summ)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return self.__key() == other.__key()


class CVEsInterface():
    def __init__(self):
        self._ver_cves = collections.defaultdict(list)

    @retry((NvdApiError, Exception), tries=3, delay=2)
    def _web_api_query(self, url, params=None):
        response = requests.get(f"{url}", params=params, timeout=10)
        if response.status_code == 200:
            response = response.json()
            return response
        else:
            msg = ""
            if "message" in response.json().keys():
                msg = response.json()["message"]
            raise NvdApiError(params, msg)

    def nist_api(self, vendor, product):
        resultsPerPage = 500

        totalResults = self.get_cves(product, vendor, resultsPerPage, 0)

        for cur_index in range(resultsPerPage, totalResults, resultsPerPage):
            self.get_cves(product, vendor, resultsPerPage, cur_index)

        return self._ver_cves

    def get_cves(self, product, vendor, resultsPerPage, cur_index):
        total_results = 0
        response = self._web_api_query("https://services.nvd.nist.gov/rest/json/cves/1.0?",
                                       params={"keyword": product, "resultsPerPage": resultsPerPage,
                                               "startIndex": cur_index})
        if response:
            self._convert_to_ranges(response["result"]["CVE_Items"], vendor, product)
            total_results = response["totalResults"]
        return total_results

    def _convert_to_ranges(self, all_cves_data, vendor, product):
        for cve_data in all_cves_data:
            cve = cve_data["cve"]['CVE_data_meta']['ID']

            if cve in self._ver_cves.keys():
                continue

            if 'configurations' not in cve_data:
                print (f'ERROR: No configurations {cve}', file = sys.stderr)
            else:
                if 'nodes' not in cve_data['configurations']:
                    print (f'ERROR: No nodes {cve}', file = sys.stderr)
                else:
                    versions = []
                    for node in cve_data['configurations']['nodes']:
                        if node['operator'] != 'OR':
                            print(f'DEBUG: No handling for OR operator in node, the following CVE needs to be implemented: {cve}', file=sys.stderr)
                        else:
                            for cpe_match in node['cpe_match']:
                                cpe_res = hashabledict()
                                if 'cpe23Uri' in cpe_match:
                                    if not f'{vendor}:{product}' in cpe_match['cpe23Uri']:
                                        continue

                                if 'versionStartIncluding' in cpe_match:
                                    cpe_res['start_including'] = cpe_match['versionStartIncluding']
                                if 'versionEndIncluding' in cpe_match:
                                    cpe_res['end_including'] = cpe_match['versionEndIncluding']
                                if 'versionStartExcluding' in cpe_match:
                                    cpe_res['start_excluding'] = cpe_match['versionStartExcluding']
                                if 'versionEndExcluding' in cpe_match:
                                    cpe_res['end_excluding'] = cpe_match['versionEndExcluding']

                                if 'cpe23Uri' in cpe_match:
                                    exact_ver = cpe_match['cpe23Uri'].partition(f'cpe:2.3:o:{vendor}:{product}:')[2].partition(':')[0]
                                    if exact_ver not in (['*', '']):
                                        cpe_res['exact'] = exact_ver

                                if cpe_res:
                                    versions.append(cpe_res)
                    if versions:
                        self._ver_cves[cve] = list(set(versions))






# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
import traceback
from ipaddress import ip_address
import sys

from commands.basecommand import BaseCommand


class FWNat(BaseCommand):
    def __init__(self):
        self.__name__ = 'FW Nat'

    def run_ssh(self, sshc):
        res = self._ssh_data_with_header(sshc, '/ip firewall nat print detail')
        sus_dns, recommendation = self.check_results_ssh(res)

        return {'raw_data': res,
                'suspicious': sus_dns,
                'recommendation': recommendation}

    def check_results_ssh(self, res):
        sus_nat = []
        recommendation = []

        try:
            for item in res:
                if (item['action'] == 'dst-nat') and ('dst-address' in item) and ('to-address' in item):
                    if (not ip_address(item['dst-address']).is_private) and (not ip_address(item['to-address']).is_private):
                        sus_nat.append(f'dst-nat rule from {item["dst-address"]} to {item["to-address"]}: both are public '
                                       f'IPs, might used for malicious activity - severity: high')
        except Exception:
            print(traceback.format_exc(), file = sys.stderr)

        return sus_nat, recommendation









